#!/usr/bin/python3
#
# mlhub - Machine Learning Model Repository
#
# A command line tool for managing machine learning models.
#
# Copyright 2018-2019 (c) Graham.Williams@togaware.com All rights reserved. 
#
# This file is part of mlhub.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the ""Software""), to deal 
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.

import json
import os
import sys
import requests
import termios
import tty
import subprocess
import re
import textwrap

# ----------------------------------------------------------------------
# Support Package Developers
# ----------------------------------------------------------------------

# Load subscription key and endpoint from file.

def load_key(path):
    key = None
    endpoint = None
    markchar = "'\" \t"
    with open(path, 'r') as file:
        for line in file:
            line = line.strip('\n')
            pair = line.split('=')
            if len(pair) == 2:
                k = pair[0].strip(markchar).lower()
                v = pair[1].strip(markchar)
                if k == 'key':
                    key = v
                elif k == 'endpoint':
                    endpoint = v
            elif not line.startswith('#'):
                line = line.strip(markchar)
                if line.startswith('http'):
                    endpoint = line
                else:
                    key = line
    return key, endpoint

# Either load key/endpoint from file or ask user and save to file.


def azkey(key_file, service="Cognitive Services", verbose=True, baseurl=False):
    """The user is asked for an Azure subscription key and endpoint. The
    provided information is saved into a file for future use. The
    contents of that file is the key and endpoint with the endpoint
    identified as starting with http. Some endpoints may include the
    full cognitive service path and so the baseurl option will strip
    to just the base name of the URL.

    a14d1234abcda4f2f6e9f565df34ef24
    https://westus2.api.cognitive.microsoft.com

    """

    key = None

    # Set up messages.
    
    prompt_key = "Please paste your {} subscription key: ".format(service)
    prompt_endpoint = "Please paste your endpoint: "

    msg_request = """\
An Azure resource is required to access this service (and to run this command).
See the README for details of a free subscription. If you have a subscription
then please paste the key and the endpoint here.
"""
    msg_found = """\
The following file has been found and is assumed to contain an Azure 
subscription key and endpoint for {}. We will load 
the file and use this information.

    {}
""".format(service, key_file)

    msg_saved = """
That information has been saved into the file:

    {}
""".format(key_file)

    # Obtain the key/endpoint.

    if os.path.isfile(key_file) and os.path.getsize(key_file) > 0:
        if verbose: print(msg_found, file=sys.stderr)
        key, endpoint = load_key(key_file)
    else:
        print(msg_request, file=sys.stderr)
        
        key      = ask_password(prompt_key)
        sys.stderr.write(prompt_endpoint)
        endpoint = input()

        if len(key) > 0 and len(endpoint) > 0:
            ofname = open(key_file, "w")
            # Use the explicit format in case endpoint has no http prefix.
            ofname.write("key={}\nendpoint={}\n".format(key, endpoint))
            ofname.close()
            print(msg_saved, file=sys.stderr)

    if baseurl:
        from urllib.parse import urlsplit
        splurl = urlsplit(endpoint)
        endpoint = splurl.scheme + "://" + splurl.netloc

    # Ensure endpoint ends in /

    if endpoint[len(endpoint) - 1] != "/": endpoint = endpoint + "/"
        
    return key, endpoint

# Simple input of password.


def ask_password(prompt=None):
    """Echo '*' for every input character. Only implements the basic I/O
    functionality and so only Backspace is supported.  No support for
    Delete, Left key, Right key and any other line editing.

    Reference: https://mail.python.org/pipermail/python-list/2011-December/615955.html
    """

    symbol = "`~!@#$%^&*()-_=+[{]}\\|;:'\",<.>/?"
    if prompt:
        sys.stderr.write(prompt)
        sys.stderr.flush()

    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    chars = []
    try:
        tty.setraw(sys.stdin.fileno())
        while True:
            c = sys.stdin.read(1)
            if c in '\n\r':  # Enter.
                break
            if c == '\003':
                raise KeyboardInterrupt
            if c == '\x7f':  # Backspace.
                if chars:
                    sys.stderr.write('\b \b')
                    sys.stderr.flush()
                    del chars[-1]
                continue
            if c.isalnum() or c in symbol:
                sys.stderr.write('*')
                sys.stderr.flush()
                chars.append(c)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        sys.stderr.write('\n')

    return ''.join(chars)

# Send a request.


def azrequest(endpoint, url, subscription_key, request_data):
    """Send anomaly detection request to the Anomaly Detector API. 

    If the request is successful, the JSON response is returned.

    Aim to generailse this to go into MLHUB to send request.
    """
    
    headers = {'Content-Type': 'application/json',
               'Ocp-Apim-Subscription-Key': subscription_key}
    
    response = requests.post(os.path.join(endpoint, url),
                             data=json.dumps(request_data),
                             headers=headers)
    
    if response.status_code == 200:
        return json.loads(response.content.decode("utf-8"))
    else:
        print(response.status_code)
        raise Exception(response.text)


def mlask(begin="", end=""):
    sys.stdout.write(begin + "Press Enter to continue: ")
    answer = input()
    sys.stdout.write(end)


def mlcat(title="", text="", delim="=", begin="", end="\n"):
    sep = delim*len(title) + "\n" if len(title) > 0 else ""
    ttl_sep = "\n" if len(title) > 0 else ""
    # Retain any extra line in the original text since fill() will
    # remove it.
    if text[-1] == "\n": end = "\n" + end
    # Split into paragraphs, fill each paragraph, convert back to a
    # list of strings, and join them together as the text to be
    # printed.
    text = "\n\n".join(list(map(textwrap.fill, text.split("\n\n"))))
    print(begin + sep + title + ttl_sep + sep + ttl_sep + text, end=end)


def mlpreview(fname,
              begin="\n",
              msg="Close the graphic window using Ctrl-w.\n",
              previewer="eog"):
    print(begin + msg)
    subprocess.Popen([previewer, fname])

# From Simon Zhao's azface package on github.


def is_url(url):
    """Check if url is a valid URL."""

    urlregex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    if re.match(urlregex, url) is not None:
        return True
    else:
        return False



#!/usr/bin/python3
#
# mlhub - Machine Learning Model Repository
#
# A command line tool for managing machine learning models.
#
# Copyright 2018 (c) Graham.Williams@togaware.com All rights reserved.
#
# This file is part of mlhub.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import collections
import logging
import os

# ------------------------------------------------------------------------
# The default ML Hub can be overridden by an environment variable or by
# the command line option --mlhub.
# ------------------------------------------------------------------------

MLHUB = "https://mlhub.ai/"
if "MLHUB" in os.environ:
    # The following adds a trailing "/" as assumed in the code.
    MLHUB = os.path.join(os.getenv("MLHUB"), "")

HUB_PATH = "pool/main/"

# ------------------------------------------------------------------------
# The MLINIT contains all of the locally installed models and configuration
# files.
# ------------------------------------------------------------------------

MLINIT = os.path.expanduser("~/.mlhub/")
if "MLINIT" in os.environ:
    # The following adds a trailing "/" as assumed in the code.
    MLINIT = os.path.join(os.getenv("MLINIT"), "")

# Cache files for bash completion.

COMPLETION_DIR = os.path.join(MLINIT, ".completion")
COMPLETION_COMMANDS = os.path.join(COMPLETION_DIR, "commands")
COMPLETION_MODELS = os.path.join(COMPLETION_DIR, "models")

COMPLETION_SCRIPT = os.path.join("bash_completion.d", "ml.bash")

# Log files

LOG_DIR = os.path.join(MLINIT, ".log")
LOG_FILE = os.path.join(LOG_DIR, "mlhub.log")

# model package cache, archive and config dir

CACHE_DIR = os.path.join(MLINIT, ".cache")
ARCHIVE_DIR = os.path.join(MLINIT, ".archive")
CONFIG_DIR = os.path.join(MLINIT, ".config")
CONFIG_FILE = "config.yaml"

# ------------------------------------------------------------------------
# Application information.
# ------------------------------------------------------------------------

APP = "mlhub"  # The application name.
APPX = "{}: ".format(APP)  # For error messages.
CMD = "ml"  # The command line tool.

EXT_MLM = ".mlm"  # Archive filename extension
EXT_AIPK = ".aipk"  # Backward compatibility

VERSION = "3.5.19"  # DO NOT MODIFY. Managed from ../Makefile.

OPTIONS = {
    # Global command line options
    #
    # All the keys in dict correspond to the keyword arguments of
    # 'argparse.ArgumentParser.add_argument()'.
    "--version": {
        "alias": ["-v"],
        "help": "display version information of ml or a package and exit.",
        "action": "store_true",
    },
    "--debug": {"help": "display debug information.", "action": "store_true"},
    "--quiet": {
        "alias": ["-q"],
        "help": "reduce noise.",
        "action": "store_true",
    },
    "--init-dir": {
        "help": "use this as the init dir instead of '{}'.".format(MLINIT)
    },
    "--mlhub": {"help": "use this ML Hub instead of '{}'.".format(MLHUB)},
    "--cmd": {
        "help": "command display name instead of '{}'.".format(CMD),
        "dest": "mlmetavar",
    },
    "--working-dir": {
        "alias": ["--wd"],
        "help": "use this as the working dir instead of '{}'/<model>.".format(
            MLINIT
        ),
    },
}

# Basic common commands

COMMANDS = {
    # -------------------------------------------------------------------
    # 'description': used in the usage by argparse and to generate next
    #                step suggestion.
    # 'suggestion': Maybe give a better suggestion instead of generating
    #               automatically from 'description'.
    # 'alias': used by argparse.
    # 'func': used by argparse to dispatch to the actual function.
    # 'next': specify the next possible commands after the command.
    #         There can be groups of next commands for different scenarios,
    #         like the command 'installed'.
    #         Next step order: 'available', 'install', 'readme', 'configure',
    #                          'commands' and commands by the order in the
    #                          model's DESCRIPTION.yaml
    # 'confirm': Maybe used for collecting user's confirmation.
    # 'argument': possible argument for the command used by argparse.
    "available": {
        "description": "list the models available from the ML Hub repository",
        "argument": {
            "--name-only": {
                "help": "list only the names",
                "action": "store_true",
            },
        },
        "alias": ["avail"],
        "usage": "  available            list the models available from the ML Hub repository",
        "func": "list_available",
        "next": ["install"],
    },
    "installed": {
        "description": "list the locally installed models",
        "argument": {
            "--name-only": {
                "help": "list only the names",
                "action": "store_true",
            },
        },
        "usage": "  installed            list the locally installed models",
        "func": "list_installed",
        "next": {
            "exist": ["configure", "readme", "commands"],
            "none": ["available", "install"],
        },
    },
    "clean": {
        "description": "remove downloaded model package files",
        "usage": "  clean                remove downloaded model package files",
        "confirm": "remove model package archive '{}' [Y/n]? ",
        "func": "remove_mlm",
    },
    "install": {
        "description": "install a named model, local model file or URL",
        "argument": {"model": {}, "-i": {"help": "SSH key path"}},
        "usage": "  install    <model>   install a named model, local model file or URL",
        "func": "install_model",
        "next": ["configure"],
    },
    # 'download': {
    #     'description': "download the actual (often large) pre-built model",
    #     'argument': {'model': {}},
    #     'func': "download_model",
    #     'next': ['readme'],
    # },
    "readme": {
        "description": "view the model's README",
        "argument": {"model": {}},
        "usage": "  readme     <model>   view the model's README",
        "func": "readme",
        "next": ["commands"],
    },
    # 'license': {
    #     'description': "display the model's LICENSE information",
    #     'argument': {'model': {}},
    #     'func': "license",
    # },
    "commands": {
        "description": "list the commands supported by the model package",
        "argument": {
            "model": {},
            "--name-only": {"help": "list names only", "action": "store_true"},
        },
        "usage": "  commands   <model>   list the commands supported by the model package",
        "func": "list_model_commands",
    },
    "configure": {
        "description": "configure the package",
        "argument": {
            "model": {"nargs": "?"},
            "-y": {"action": "store_true", "help": 'the same as "--yes"'},
            "--yes": {
                "action": "store_true",
                "help": 'assume "yes" as answer to all prompts',
            },
            "-i": {"help": "SSH key path"},
        },
        "usage": "  configure [<model>]  configure ml or the model's dependencies",
        "func": "configure_model",
        "next": ["readme"],
    },
    "remove": {
        "description": "remove a model or remove all models",
        "argument": {"model": {"nargs": "?"}},
        "usage": "  remove    [<model>]  remove a model or remove all models",
        "func": "remove_model",
        "next": ["installed", "install"],
    },
    # 'demo': {
    #     'description': "run the model's demonstration",
    #     'alias': ['print', 'display', 'score', 'rebuild'],
    #     'argument': {
    #         'model': {},
    #         'param': {'nargs': "*"}
    #     },
    #     'usage': "  demo       <model>   demonstrate the model in action",
    #     'func': "dispatch",
    # },
    # 'donate': {
    #     'description': "consider a donation to the author.",
    #     'argument': {'model': {}},
    #     'func': "donate",
    # },
}

COMMANDS_USAGE = collections.OrderedDict()
for cmd in collections.OrderedDict(COMMANDS):
    argname = ""
    if "argument" in COMMANDS[cmd]:
        for k, v in COMMANDS[cmd]["argument"].items():
            if not k.startswith("-"):
                argname = "<" + k + ">"
                if "nargs" in v and v["nargs"] == "?":
                    argname = "[" + argname + "]"
                break
    usage = "  {:10s}{:^9s}  {}".format(
        cmd, argname, COMMANDS[cmd]["description"]
    )
    COMMANDS_USAGE[cmd] = usage

USAGE = """Usage: {{}} [<options>] <command> [<command options>] [<model>]

Access machine learning models from the ML Hub.

Global commands:

{}

{}

The ML Hub repository is '{{}}'.

Models are installed into '{{}}'.

This is version {{}} of {{}}.

To ensure the pre-requisites are installed and for a better experience
with tab completion the sys admin can run the command:

  $ ml configure

The user can then run the following for tab completion:

  $ source /etc/bash_completion.d/ml.bash

List the available models from the repository with:

  $ ml available
""".format(
    "\n".join(list(COMMANDS_USAGE.values())[:3]),
    "\n".join(list(COMMANDS_USAGE.values())[3:]),
)

# ------------------------------------------------------------------------
# File names
# ------------------------------------------------------------------------

README = "README.txt"

MLHUB_YAML = "MLHUB.yaml"

DESC_YAML = "DESCRIPTION.yaml"
META_YAML = "Packages.yaml"

DESC_YML = "DESCRIPTION.yml"
META_YML = "Packages.yml"

# ------------------------------------------------------------------------
# Debugging
# ------------------------------------------------------------------------

LOG_FILE_LEVEL = logging.DEBUG
LOG_FILE_FORMAT = "%(asctime)s - %(name)s - %(levelname)s: %(message)s"
LOG_CONSOLE_FORMAT = "--> %(name)s - %(levelname)s: %(message)s"
LOG_NOT_QUIET = {"quiet": False}
LOG_QUIET = {"quiet": True}

# ------------------------------------------------------------------------
# Model package config entry keys
# ------------------------------------------------------------------------

CONDA_ENV_NAME = "conda_env_name"  # Conda environment name
WORKING_DIR = "working_dir"  # Model's working dir
PYTHON_PATH = "python_path"  # python path, such as /usr/bin/python3
PIP_PATH = "pip_path"  # pip path, such as /usr/bin/pip3
SYS_PYTHON_PKG_USAGE = (
    "sys_python_pkg_usage"  # Whether system python packages installed
)


# ------------------------------------------------------------------------
# Command binary
# ------------------------------------------------------------------------

BASH_CMD = "/bin/bash"

R_CMD = "/usr/bin/R"
RSCRIPT_CMD = "/usr/bin/Rscript"

SYS_PYTHON_CMD = "/usr/bin/python3"
SYS_PIP_CMD = "/usr/bin/pip3"


# ------------------------------------------------------------------------
# Messages
# ------------------------------------------------------------------------

MSG_INCOMPATIBLE_PYTHON_ENV = """
WARNING: MLHub is not installed in system's site package directory!
The "{}" MLHub package requires system Python packages, which may not correctly
be included in the its Python search path.
To solve this problem, please make sure to install MLHub using:
    $ pip uninstall -y mlhub  # or $ pip3 uninstall -y mlhub
    $ sudo apt-get install -y python3-pip
    $ /usr/bin/pip3 install mlhub
followed by re-login.
Or for advanced users, you could manually `pip install` those required Python
packages.
"""
